from typing import Tuple, Type

from django.db.models import Model


def reverse_relationship(
    relationship_expr: str, root_model: Type[Model]
) -> Tuple[str, Type[Model]]:
    """
    Reverses a relationship expression relative to root_model.

    Args:
        relationship_expr: The Django relationship string, with underscores to
            represent relationship traversal.
        root_model: The model to which relationship_expr is relative.

    Returns:
        str: The django relationship string in reverse, so from the last joined
            relationship back to the root model.
        Type[Model]: The model to which the returned expression is relative.
    """
    relation_steps = relationship_expr.split("__")

    related_model = root_model
    path_to_outerref_parts = []
    for step in relation_steps:
        related_field = related_model._meta.get_field(step)
        related_model = related_field.related_model
        path_to_outerref_parts.append(related_field.remote_field.name)

    path_to_outerref = "__".join(reversed(path_to_outerref_parts))

    return (path_to_outerref, related_model)



from django.db.models import CharField, Lookup, Subquery, fields, functions


@fields.Field.register_lookup
class NotEqual(Lookup):
    """https://docs.djangoproject.com/en/2.2/howto/custom-lookups/"""

    lookup_name = "ne"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s <> %s" % (lhs, rhs), params


CharField.register_lookup(functions.Length)
CharField.register_lookup(functions.Upper)
CharField.register_lookup(functions.Lower)
CharField.register_lookup(functions.Trim)


class _AnyAll(Subquery):
    template = ""

    def __init__(self, queryset, negated=False, **kwargs):
        # As a performance optimization, remove ordering since ~ doesn't
        # care about it, just whether or not a row matches.
        queryset = queryset.order_by()
        self.negated = negated
        super().__init__(queryset, **kwargs)

    def __invert__(self):
        clone = self.copy()
        clone.negated = not self.negated
        return clone

    def __repr__(self) -> str:
        return self.template % {"subquery": self.queryset.query}

    def as_sql(self, compiler, connection, template=None, **extra_context):
        sql, params = super().as_sql(compiler, connection, template, **extra_context)
        # if self.negated:
        #     sql = "NOT {}".format(sql)
        return sql, params

    def select_format(self, compiler, sql, params):
        return sql, params


class Any(_AnyAll):
    template = "ANY(%(subquery)s)"


class All(_AnyAll):
    template = "ALL(%(subquery)s)"



from typing import List

from sqlalchemy.sql.expression import ClauseElement, Select

from odata_query.grammar import ODataLexer, ODataParser  # type: ignore

from .sqlalchemy_clause import AstToSqlAlchemyClauseVisitor


def _get_joined_attrs(query: Select) -> List[str]:
    return [str(join[0]) for join in query._setup_joins]


def apply_odata_query(query: ClauseElement, odata_query: str) -> ClauseElement:
    """
    Shorthand for applying an OData query to a SQLAlchemy query.

    Args:
        query: SQLAlchemy query to apply the OData query to.
        odata_query: OData query string.
    Returns:
        ClauseElement: The modified query
    """
    lexer = ODataLexer()
    parser = ODataParser()
    model = query.column_descriptions[0]["entity"]

    ast = parser.parse(lexer.tokenize(odata_query))
    transformer = AstToSqlAlchemyClauseVisitor(model)
    where_clause = transformer.visit(ast)

    for j in transformer.join_relationships:
        if str(j) not in _get_joined_attrs(query):
            query = query.join(j)

    return query.filter(where_clause)



from . import ast
from .rewrite import IdentifierStripper


def expression_relative_to_identifier(
    identifier: ast.Identifier, expression: ast._Node
) -> ast._Node:
    """
    Shorthand for the :class:`IdentifierStripper`.

    Args:
        identifier: Identifier to strip from ``expression``.
        expression: Expression to strip the ``identifier`` from.

    Returns:
        The ``expression`` relative to the ``identifier``.
    """
    stripper = IdentifierStripper(identifier)
    result = stripper.visit(expression)
    return result



from sqlalchemy.sql.functions import GenericFunction
from sqlalchemy.types import Integer, String


class strpos(GenericFunction):
    type = Integer


class substr(GenericFunction):
    type = String


class lower(GenericFunction):
    type = String


class upper(GenericFunction):
    type = String


class ltrim(GenericFunction):
    type = String


class rtrim(GenericFunction):
    type = String


class ceil(GenericFunction):
    type = Integer


class floor(GenericFunction):
    type = Integer


class round(GenericFunction):
    type = Integer



class ODataException(Exception):
    """
    Base class for all exceptions in this library.
    """

    pass


class ODataSyntaxError(ODataException):
    """
    Base class for syntax errors.
    """

    pass


class TokenizingException(ODataSyntaxError):
    """
    Thrown when the lexer cannot tokenize the query.
    """

    def __init__(self, token):
        self.token = token
        super().__init__(f"Failed to tokenize at: {token}")


class ParsingException(ODataSyntaxError):
    """
    Thrown when the parser cannot parse the query.
    """

    def __init__(self, token, eof: bool = False):
        self.token = token
        self.eof = eof
        super().__init__(f"Failed to parse at: {token}")


class FunctionCallException(ODataException):
    """
    Base class for errors in function calls.
    """

    pass


class UnknownFunctionException(FunctionCallException):
    """
    Thrown when the parser encounters an undefined function call.
    """

    def __init__(self, function_name: str):
        self.function_name = function_name
        super().__init__(f"Unknown function: '{function_name}'")


class ArgumentCountException(FunctionCallException):
    """
    Thrown when the parser encounters a function called with a wrong number
    of arguments.
    """

    def __init__(
        self, function_name: str, exp_min_args: int, exp_max_args: int, given_args: int
    ):
        self.function_name = function_name
        self.exp_min_args = exp_min_args
        self.exp_max_args = exp_max_args
        self.n_args_given = given_args
        if exp_min_args != exp_max_args:
            super().__init__(
                f"Function '{function_name}' takes between {exp_min_args} and "
                f"{exp_max_args} arguments. {given_args} given."
            )
        else:
            super().__init__(
                f"Function '{function_name}' takes {exp_min_args} arguments. "
                f"{given_args} given."
            )


class UnsupportedFunctionException(FunctionCallException):
    """
    Thrown when a function is used that is not implemented yet.
    """

    def __init__(self, function_name: str):
        self.function_name = function_name
        super().__init__(f"Function '{function_name}' is not implemented yet.")


class ArgumentTypeException(FunctionCallException):
    """
    Thrown when a function is called with argument of the wrong type.
    """

    def __init__(
        self,
        function_name: str = None,
        expected_type: str = None,
        actual_type: str = None,
    ):
        self.function_name = function_name
        self.expected_type = expected_type
        self.actual_type = actual_type

        if function_name:
            message = f"Unsupported or invalid type for function or operator '{function_name}'"
        else:
            message = "Invalid argument type for function or operator."
        if expected_type:
            message += " Expected {expected_type}"
            if actual_type:
                message += ", got {actual_type}"

        super().__init__(message)


class TypeException(ODataException):
    """
    Thrown when doing an invalid operation on a value.
    E.g. `10 gt null` or `~date()`
    """

    def __init__(self, operation: str, value: str):
        self.operation = operation
        self.value = value
        super().__init__(f"Cannot apply '{operation}' to '{value}'")


class ValueException(ODataException):
    """
    Thrown when a value has an invalid value, such as an invalid datetime.
    """

    def __init__(self, value):
        self.value = value
        super().__init__(f"Invalid value: {value}")


class InvalidFieldException(ODataException):
    """
    Thrown when a field mentioned in a query does not exist.
    """

    def __init__(self, field_name: str):
        self.field_name = field_name
        super().__init__(f"Invalid field: {field_name}")



import datetime as dt
import operator
import uuid
from typing import Any, Callable, Dict, List, Optional, Type

from dateutil.parser import isoparse
from django.db.models import Exists, F, Model, OuterRef, Q, Value, functions
from django.db.models.expressions import Expression

from odata_query import ast, exceptions as ex, typing, utils, visitor

from .utils import reverse_relationship

COMPARISON_INVERT = {
    ast.Eq: ast.Eq,
    ast.NotEq: ast.NotEq,
    ast.Lt: ast.Gt,
    ast.LtE: ast.GtE,
    ast.Gt: ast.Lt,
    ast.GtE: ast.LtE,
}


class AstToDjangoQVisitor(visitor.NodeVisitor):
    """
    :class:`NodeVisitor` that transforms an :term:`AST` into a Django Q
    filter object.

    Args:
        root_model: The root model of the query.
    """

    def __init__(self, root_model: Type[Model]):
        self.root_model = root_model
        self.queryset_annotations: Dict[str, Expression] = {}

    def visit_Identifier(self, node: ast.Identifier) -> F:
        ":meta private:"
        return F(node.name)

    def visit_Attribute(self, node: ast.Attribute) -> F:
        ":meta private:"
        owner = self.visit(node.owner)
        full_id = owner.name + "__" + node.attr
        return F(full_id)

    def visit_Null(self, node: ast.Null) -> str:
        ":meta private:"
        raise NotImplementedError("Should not be reached")

    def visit_Integer(self, node: ast.Integer) -> Value:
        ":meta private:"
        return Value(int(node.val))

    def visit_Float(self, node: ast.Float) -> Value:
        ":meta private:"
        return Value(float(node.val))

    def visit_Boolean(self, node: ast.Boolean) -> Value:
        ":meta private:"
        return Value(node.val == "true")

    def visit_String(self, node: ast.String) -> Value:
        ":meta private:"
        return Value(node.val)

    def visit_Date(self, node: ast.Date) -> Value:
        ":meta private:"
        try:
            return Value(dt.date.fromisoformat(node.val))
        except ValueError:
            raise ex.ValueException(node.val)

    def visit_DateTime(self, node: ast.DateTime) -> Value:
        ":meta private:"
        try:
            return Value(isoparse(node.val))
        except ValueError:
            raise ex.ValueException(node.val)

    def visit_Time(self, node: ast.Time) -> Value:
        ":meta private:"
        try:
            return Value(dt.time.fromisoformat(node.val))
        except ValueError:
            raise ex.ValueException(node.val)

    def visit_Duration(self, node: ast.Duration) -> Value:
        ":meta private:"
        sign, days, hours, minutes, seconds = node.unpack()
        td = dt.timedelta(
            days=float(days or 0),
            hours=float(hours or 0),
            minutes=float(minutes or 0),
            seconds=float(seconds or 0),
        )
        if sign and sign == "-":
            td = -1 * td
        return Value(td)

    def visit_GUID(self, node: ast.GUID) -> Value:
        ":meta private:"
        return uuid.UUID(node.val)

    def visit_List(self, node: ast.List) -> List:
        ":meta private:"
        return [self.visit(n) for n in node.val]

    def visit_Add(self, node: ast.Add) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.add

    def visit_Sub(self, node: ast.Sub) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.sub

    def visit_Mult(self, node: ast.Mult) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.mul

    def visit_Div(self, node: ast.Div) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.truediv

    def visit_Mod(self, node: ast.Mod) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.mod

    def visit_BinOp(self, node: ast.BinOp) -> Any:
        ":meta private:"
        # Left or right can be an Identifier, in which case it needs to be
        # wrapped in F()
        left = self.visit(node.left)
        right = self.visit(node.right)
        op = self.visit(node.op)

        return op(left, right)

    def visit_Eq(self, node: ast.Eq) -> str:
        ":meta private:"
        return "__exact"

    def visit_NotEq(self, node: ast.NotEq) -> str:
        ":meta private:"
        return "__ne"

    def visit_Lt(self, node: ast.Lt) -> str:
        ":meta private:"
        return "__lt"

    def visit_LtE(self, node: ast.LtE) -> str:
        ":meta private:"
        return "__lte"

    def visit_Gt(self, node: ast.Gt) -> str:
        ":meta private:"
        return "__gt"

    def visit_GtE(self, node: ast.GtE) -> str:
        ":meta private:"
        return "__gte"

    def visit_In(self, node: ast.In) -> str:
        ":meta private:"
        return "__in"

    def visit_Compare(self, node: ast.Compare) -> Q:
        ":meta private:"
        # Special case: comparison to NULL => isnull=True/False
        # Should not be wrapped with Value(True/False)
        # See: https://github.com/django/django/blob/0aacbdcf27b258387643b033352e99e6103abda8/django/db/models/lookups.py#L515
        if isinstance(node.right, ast.Null):
            lhs = self._attempt_keywordify(node.left)
            if not lhs:
                raise ex.TypeException(node.comparator.__class__.__name__, "null")
            q_keyword = lhs + "__isnull"
            if isinstance(node.comparator, ast.Eq):
                return Q(**{q_keyword: True})
            elif isinstance(node.comparator, ast.NotEq):
                return Q(**{q_keyword: False})
            else:
                raise ex.TypeException(node.comparator.__class__.__name__, "null")

        # Need an identifier on any side to make a Django Q:
        keyword = self._attempt_keywordify(node.left)
        if not keyword:
            # No keyword on the left, try the right:
            keyword = self._attempt_keywordify(node.right)
            if keyword:
                # Keyword on right, flip the comparison so it's left now:
                node = self._flip_comparison(node)
            else:
                # No keywords at all, cannot continue:
                raise ex.TypeException(
                    node.comparator.__class__.__name__, str(node.left)
                )

        q_keyword = keyword + self.visit(node.comparator)
        query = Q(**{q_keyword: self.visit(node.right)})

        return query

    def visit_And(self, node: ast.And) -> Callable[[Q, Q], Q]:
        ":meta private:"
        return operator.and_

    def visit_Or(self, node: ast.Or) -> Callable[[Q, Q], Q]:
        ":meta private:"
        return operator.or_

    def visit_BoolOp(self, node: ast.BoolOp) -> Q:
        ":meta private:"
        left = self.visit(node.left)
        right = self.visit(node.right)

        if isinstance(left, (F, Value)):
            raise ex.TypeException(node.op.__class__.__name__, left)
        if isinstance(right, (F, Value)):
            raise ex.TypeException(node.op.__class__.__name__, right)

        op = self.visit(node.op)

        return op(left, right)

    def visit_Not(self, node: ast.Not) -> Callable[[Q], Q]:
        ":meta private:"
        return operator.invert

    def visit_UnaryOp(self, node: ast.UnaryOp) -> str:
        ":meta private:"
        mod = self.visit(node.op)
        val = self.visit(node.operand)

        try:
            return mod(val)
        except TypeError:
            raise ex.TypeException(node.op.__class__.__name__, val)

    def visit_Call(self, node: ast.Call) -> Q:
        ":meta private:"
        try:
            q_gen = getattr(self, "djangofunc_" + node.func.name.lower())
        except AttributeError:
            raise ex.UnsupportedFunctionException(node.func.name)

        return q_gen(*node.args)

    def visit_CollectionLambda(self, node: ast.CollectionLambda) -> Q:
        ":meta private:"
        # NOTE: The initial implementation translated to SQL's ANY/ALL keywords,
        # but those behave very differently from OData's any/all keywords!

        owner_path = self._attempt_keywordify(node.owner)
        if not owner_path:
            raise ex.TypeException("lambda_expression", str(node.owner))

        path_to_outerref, related_model = reverse_relationship(
            owner_path, self.root_model
        )
        subquery = related_model.objects.filter(Q(**{path_to_outerref: OuterRef("pk")}))
        # .values(related_field.remote_field.name)

        if node.lambda_:
            # For the lambda, we want to strip the identifier off, because
            # we will execute this as a subquery in the wanted model's context.
            subq_ast = utils.expression_relative_to_identifier(
                node.lambda_.identifier, node.lambda_.expression
            )
            subq_transformer = self.__class__(related_model)
            subquery_filter = subq_transformer.visit(subq_ast)
        else:
            subquery_filter = None

        if isinstance(node.operator, ast.Any):
            # If ANY item should match in the subquery, we can use EXISTS():
            if subquery_filter:
                subquery = subquery.filter(subquery_filter)
            return Exists(subquery)

        elif isinstance(node.operator, ast.All):
            # If ALL items in the collection must match, we invert the condition and use NOT EXISTS():
            if subquery_filter:
                subquery = subquery.filter(~subquery_filter)
            return Exists(subquery, negated=True)

        else:
            raise NotImplementedError()

    def djangofunc_contains(self, field: ast._Node, substr: ast._Node) -> Q:
        ":meta private:"
        return self._substr_function(field, substr, "contains")

    def djangofunc_startswith(self, field: ast._Node, substr: ast._Node) -> Q:
        ":meta private:"
        return self._substr_function(field, substr, "startswith")

    def djangofunc_endswith(self, field: ast._Node, substr: ast._Node) -> Q:
        ":meta private:"
        return self._substr_function(field, substr, "endswith")

    def djangofunc_length(self, arg: ast._Node) -> functions.Length:
        ":meta private:"
        return functions.Length(self.visit(arg))

    def djangofunc_concat(self, *args: ast._Node) -> functions.Concat:
        ":meta private:"
        return functions.Concat(*[self.visit(arg) for arg in args])

    def djangofunc_indexof(
        self, first: ast._Node, second: ast._Node
    ) -> functions.StrIndex:
        ":meta private:"
        # Subtract 1 because OData is 0-indexed while SQL is 1-indexed
        return functions.StrIndex(self.visit(first), self.visit(second)) - 1

    def djangofunc_substring(
        self, fullstr: ast._Node, index: ast._Node, nchars: ast._Node = None
    ) -> functions.Substr:
        ":meta private:"
        # Add 1 because OData is 0-indexed while SQL is 1-indexed
        return functions.Substr(
            self.visit(fullstr),
            self.visit(index) + 1,
            self.visit(nchars) if nchars else None,
        )

    def djangofunc_matchespattern(self, field: ast._Node, pattern: ast._Node) -> Q:
        ":meta private:"
        lhs = self._attempt_keywordify(field)
        if not lhs:
            raise ex.ArgumentTypeException("matchespattern")
        q_keyword = lhs + "__regex"
        pattern = self.visit(pattern)

        return Q(**{q_keyword: pattern})

    def djangofunc_tolower(self, field: ast._Node) -> functions.Lower:
        ":meta private:"
        return functions.Lower(self.visit(field))

    def djangofunc_toupper(self, field: ast._Node) -> functions.Upper:
        ":meta private:"
        return functions.Upper(self.visit(field))

    def djangofunc_trim(self, field: ast._Node) -> functions.Trim:
        ":meta private:"
        return functions.Trim(self.visit(field))

    def djangofunc_date(self, field: ast._Node) -> functions.TruncDate:
        ":meta private:"
        return functions.TruncDate(self.visit(field))

    def djangofunc_day(self, field: ast._Node) -> functions.ExtractDay:
        ":meta private:"
        return functions.ExtractDay(self.visit(field))

    def djangofunc_hour(self, field: ast._Node) -> functions.ExtractHour:
        ":meta private:"
        return functions.ExtractHour(self.visit(field))

    def djangofunc_minute(self, field: ast._Node) -> functions.ExtractMinute:
        ":meta private:"
        return functions.ExtractMinute(self.visit(field))

    def djangofunc_month(self, field: ast._Node) -> functions.ExtractMonth:
        ":meta private:"
        return functions.ExtractMonth(self.visit(field))

    def djangofunc_now(self) -> functions.Now:
        ":meta private:"
        return functions.Now()

    def djangofunc_second(self, field: ast._Node) -> functions.ExtractSecond:
        ":meta private:"
        return functions.ExtractSecond(self.visit(field))

    def djangofunc_time(self, field: ast._Node) -> functions.TruncTime:
        ":meta private:"
        return functions.TruncTime(self.visit(field))

    def djangofunc_year(self, field: ast._Node) -> functions.ExtractYear:
        ":meta private:"
        return functions.ExtractYear(self.visit(field))

    def djangofunc_ceiling(self, field: ast._Node) -> functions.Ceil:
        ":meta private:"
        return functions.Ceil(self.visit(field))

    def djangofunc_floor(self, field: ast._Node) -> functions.Floor:
        ":meta private:"
        return functions.Floor(self.visit(field))

    def djangofunc_round(self, field: ast._Node) -> functions.Round:
        ":meta private:"
        return functions.Round(self.visit(field))

    def _substr_function(
        self, field: ast._Node, substr: ast._Node, django_func: str
    ) -> Q:
        ":meta private:"
        typing.typecheck(field, (ast.Identifier, ast.String), "field")
        typing.typecheck(substr, ast.String, "substring")

        lhs = self._attempt_keywordify(field)
        if not lhs:
            raise ex.ArgumentTypeException(django_func)
        q_keyword = lhs + "__" + django_func
        substring = self.visit(substr)

        return Q(**{q_keyword: substring})

    def _attempt_keywordify(self, node: ast._Node) -> Optional[str]:
        ":meta private:"
        if isinstance(node, ast._Literal):
            return None

        res = self.visit(node)
        if isinstance(res, F):
            return res.name

        if (
            hasattr(res, "lookup_name")
            and hasattr(res, "lhs")
            and hasattr(res.lhs, "name")
        ):
            return res.lhs.name + "__" + res.lookup_name

        # Attempt to make this a QS annotation (SQL alias):
        if isinstance(res, Expression):
            identity = self._gen_annotation_name(res)
            self.queryset_annotations[identity] = res
            return identity

        return None

    def _gen_annotation_name(self, expr: Expression) -> str:
        ":meta private:"
        if hasattr(expr, "name"):
            return expr.name
        elif hasattr(expr, "value"):
            return str(expr.value)

        func_name = expr.__class__.__name__

        try:
            args = expr.get_source_expressions()
        except AttributeError:
            args = []

        args_str = [self._gen_annotation_name(a) for a in args]

        return (
            "_".join([func_name] + args_str)
            .replace(" ", "_")
            .replace(",", "")
            .replace(":", "_")
            .lower()
        )

    @staticmethod
    def _flip_comparison(comp: ast.Compare) -> ast.Compare:
        """
        Flip a comparison left-to-right. E.g.: (4 > version_id) becomes (version_id < 4)

        :meta private:
        """
        new_op = COMPARISON_INVERT[type(comp.comparator)]()
        new_left = comp.right
        new_right = comp.left

        return ast.Compare(new_op, new_left, new_right)



import re
from dataclasses import dataclass
from typing import List as ListType, Optional, Tuple

DURATION_PATTERN = re.compile(r"([+-])?P(\d+D)?(?:T(\d+H)?(\d+M)?(\d+(?:\.\d+)?S)?)?")


@dataclass(frozen=True)
class _Node:
    pass


@dataclass(frozen=True)
class Identifier(_Node):
    name: str


@dataclass(frozen=True)
class Attribute(_Node):
    owner: _Node
    attr: str


###############################################################################
# Literals
###############################################################################
@dataclass(frozen=True)
class _Literal(_Node):
    pass


@dataclass(frozen=True)
class Null(_Literal):
    pass


@dataclass(frozen=True)
class Integer(_Literal):
    val: str


@dataclass(frozen=True)
class Float(_Literal):
    val: str


@dataclass(frozen=True)
class Boolean(_Literal):
    val: str


@dataclass(frozen=True)
class String(_Literal):
    val: str


@dataclass(frozen=True)
class Date(_Literal):
    val: str


@dataclass(frozen=True)
class Time(_Literal):
    val: str


@dataclass(frozen=True)
class DateTime(_Literal):
    val: str


@dataclass(frozen=True)
class Duration(_Literal):
    val: str

    def unpack(
        self,
    ) -> Tuple[
        Optional[str], Optional[str], Optional[str], Optional[str], Optional[str]
    ]:
        """
        Returns:
            ``(sign, days, hours, minutes, seconds)``
        """

        match = DURATION_PATTERN.fullmatch(self.val)
        if not match:
            raise ValueError(f"Could not unpack Duration with value {self.val}")

        sign, days, hours, minutes, seconds = match.groups()

        _days = days[:-1] if days else None
        _hours = hours[:-1] if hours else None
        _minutes = minutes[:-1] if minutes else None
        _seconds = seconds[:-1] if seconds else None

        return sign, _days, _hours, _minutes, _seconds


@dataclass(frozen=True)
class GUID(_Literal):
    val: str


@dataclass(frozen=True)
class List(_Literal):
    val: ListType[_Literal]


###############################################################################
# Arithmetic
###############################################################################
@dataclass(frozen=True)
class _BinOpToken(_Node):
    pass


@dataclass(frozen=True)
class Add(_BinOpToken):
    pass


@dataclass(frozen=True)
class Sub(_BinOpToken):
    pass


@dataclass(frozen=True)
class Mult(_BinOpToken):
    pass


@dataclass(frozen=True)
class Div(_BinOpToken):
    pass


@dataclass(frozen=True)
class Mod(_BinOpToken):
    pass


@dataclass(frozen=True)
class BinOp(_Node):
    op: _BinOpToken
    left: _Node
    right: _Node


###############################################################################
# Comparison
###############################################################################
@dataclass(frozen=True)
class _Comparator(_Node):
    pass


@dataclass(frozen=True)
class Eq(_Comparator):
    pass


@dataclass(frozen=True)
class NotEq(_Comparator):
    pass


@dataclass(frozen=True)
class Lt(_Comparator):
    pass


@dataclass(frozen=True)
class LtE(_Comparator):
    pass


@dataclass(frozen=True)
class Gt(_Comparator):
    pass


@dataclass(frozen=True)
class GtE(_Comparator):
    pass


@dataclass(frozen=True)
class In(_Comparator):
    pass


@dataclass(frozen=True)
class Compare(_Node):
    comparator: _Comparator
    left: _Node
    right: _Node


###############################################################################
# Boolean ops
###############################################################################
@dataclass(frozen=True)
class _BoolOpToken(_Node):
    pass


@dataclass(frozen=True)
class And(_BoolOpToken):
    pass


@dataclass(frozen=True)
class Or(_BoolOpToken):
    pass


@dataclass(frozen=True)
class BoolOp(_Node):
    op: _BoolOpToken
    left: _Node
    right: _Node


###############################################################################
# Unary ops
###############################################################################
@dataclass(frozen=True)
class _UnaryOpToken(_Node):
    pass


@dataclass(frozen=True)
class Not(_UnaryOpToken):
    pass


@dataclass(frozen=True)
class USub(_UnaryOpToken):
    pass


@dataclass(frozen=True)
class UnaryOp(_Node):
    op: _UnaryOpToken
    operand: _Node


###############################################################################
# Function calls
###############################################################################
@dataclass(frozen=True)
class Call(_Node):
    func: Identifier
    args: ListType[_Node]


###############################################################################
# Collections
###############################################################################
@dataclass(frozen=True)
class _CollectionOperator(_Node):
    pass


@dataclass(frozen=True)
class Any(_CollectionOperator):
    pass


@dataclass(frozen=True)
class All(_CollectionOperator):
    pass


@dataclass(frozen=True)
class Lambda(_Node):
    identifier: Identifier
    expression: _Node


@dataclass(frozen=True)
class CollectionLambda(_Node):
    owner: _Node
    operator: _CollectionOperator
    lambda_: Optional[Lambda]



import logging
import re
from typing import Optional

from odata_query import ast, exceptions, typing, visitor

log = logging.getLogger(__name__)

UNSAFE_CHARS = re.compile(r"[^a-zA-Z0-9_]")


def clean_athena_identifier(identifier: str) -> str:
    """
    Cleans an Athena identifier so it passes the following validation rules:

    - Table names and table column names in Athena must be lowercase
    - Athena table, view, database, and column names allow only underscore special characters
    - Names should be quoted or backticked when starting with a number or underscore

    Source: https://docs.aws.amazon.com/athena/latest/ug/tables-databases-columns-names.html
    """
    id_new = identifier.lower()
    id_new = UNSAFE_CHARS.sub("_", id_new)

    first_char = id_new[0]
    if first_char == "_" or first_char.isdigit():
        log.warning(
            "Athena identifier '%s' starts with an unsafe character, "
            "which might reduce compatibility with certain BI tools.",
            identifier,
        )

    return id_new


class AstToAthenaSqlVisitor(visitor.NodeVisitor):
    """
    :class:`NodeVisitor` that transforms an :term:`AST` into an Athena SQL
    ``WHERE`` clause.

    Args:
        table_alias: Optional alias for the root table.
    """

    def __init__(self, table_alias: Optional[str] = None):
        super().__init__()
        self.table_alias = table_alias

    def visit_Identifier(self, node: ast.Identifier) -> str:
        ":meta private:"
        # Double quotes for column names acc SQL Standard
        sql_id = f'"{clean_athena_identifier(node.name)}"'

        if self.table_alias:
            sql_id = f'"{self.table_alias}".' + sql_id

        return sql_id

    def visit_Null(self, node: ast.Null) -> str:
        ":meta private:"
        return "NULL"

    def visit_Integer(self, node: ast.Integer) -> str:
        ":meta private:"
        return node.val

    def visit_Float(self, node: ast.Float) -> str:
        ":meta private:"
        return node.val

    def visit_Boolean(self, node: ast.Boolean) -> str:
        ":meta private:"
        return node.val.upper()

    def visit_String(self, node: ast.String) -> str:
        ":meta private:"
        # Replace single quotes with double single-quotes acc SQL standard:
        val = node.val.replace("'", "''")
        # Wrap in single quotes for string constants acc SQL Standard
        return f"'{val}'"

    def visit_Date(self, node: ast.Date) -> str:
        ":meta private:"
        # Single quotes for date constants acc SQL Standard
        return f"DATE '{node.val}'"

    def visit_DateTime(self, node: ast.DateTime) -> str:
        ":meta private:"
        # Single quotes for datetime constants acc SQL Standard
        return f"from_iso8601_timestamp('{node.val}')"

    def visit_Duration(self, node: ast.Duration) -> str:
        ":meta private:"
        sign, days, hours, minutes, seconds = node.unpack()
        millis: Optional[str]

        if seconds:
            if "." in seconds:
                seconds, millis = seconds.split(".")
                # Get exactly length 3 millis:
                millis = "{:0<3.3}".format(millis)
            else:
                millis = None
        else:
            millis = None

        sign = sign or ""
        intervals = []
        if days:
            intervals.append(f"INTERVAL '{days}' DAY")
        if hours:
            intervals.append(f"INTERVAL '{hours}' HOUR")
        if minutes:
            intervals.append(f"INTERVAL '{minutes}' MINUTE")
        if seconds:
            intervals.append(f"INTERVAL '{seconds}' SECOND")
        if millis:
            intervals.append(f"INTERVAL '{millis}' MILLISECOND")

        if len(intervals) == 0:
            # Shouldn't occur but whatever
            return ""
        if len(intervals) == 1:
            return f"{sign}{intervals[0]}"
        if len(intervals) > 1:
            interval = " + ".join(intervals)
            return f"{sign}({interval})"

        # Make Quality checks happy:
        raise Exception("This code is never reachable...")

    def visit_GUID(self, node: ast.GUID) -> str:
        ":meta private:"
        return f"'{node.val}'"

    def visit_List(self, node: ast.List) -> str:
        ":meta private:"
        options = ", ".join(self.visit(n) for n in node.val)
        return f"({options})"

    def visit_Add(self, node: ast.Add) -> str:
        ":meta private:"
        return "+"

    def visit_Sub(self, node: ast.Sub) -> str:
        ":meta private:"
        return "-"

    def visit_Mult(self, node: ast.Mult) -> str:
        ":meta private:"
        return "*"

    def visit_Div(self, node: ast.Div) -> str:
        ":meta private:"
        return "/"

    def visit_Mod(self, node: ast.Mod) -> str:
        ":meta private:"
        return "%"

    def visit_BinOp(self, node: ast.BinOp) -> str:
        ":meta private:"
        left = self.visit(node.left)
        right = self.visit(node.right)
        op = self.visit(node.op)

        return f"{left} {op} {right}"

    def visit_Eq(self, node: ast.Eq) -> str:
        ":meta private:"
        return "="

    def visit_NotEq(self, node: ast.NotEq) -> str:
        ":meta private:"
        return "!="

    def visit_Lt(self, node: ast.Lt) -> str:
        ":meta private:"
        return "<"

    def visit_LtE(self, node: ast.LtE) -> str:
        ":meta private:"
        return "<="

    def visit_Gt(self, node: ast.Gt) -> str:
        ":meta private:"
        return ">"

    def visit_GtE(self, node: ast.GtE) -> str:
        ":meta private:"
        return ">="

    def visit_In(self, node: ast.In) -> str:
        ":meta private:"
        return "IN"

    def visit_Compare(self, node: ast.Compare) -> str:
        ":meta private:"
        left = self.visit(node.left)
        right = self.visit(node.right)
        comparator = self.visit(node.comparator)

        # In case of a subexpression, wrap it in parentheses
        if isinstance(node.left, (ast.BoolOp, ast.Compare)):
            left = f"({left})"
        if isinstance(node.right, (ast.BoolOp, ast.Compare)):
            right = f"({right})"

        #  'eq/ne null' should become 'IS (NOT) NULL' instead of '(!)= NULL'
        if isinstance(node.right, ast.Null):
            if isinstance(node.comparator, ast.Eq):
                comparator = "IS"
            elif isinstance(node.comparator, ast.NotEq):
                comparator = "IS NOT"

        return f"{left} {comparator} {right}"

    def visit_And(self, node: ast.And) -> str:
        ":meta private:"
        return "AND"

    def visit_Or(self, node: ast.Or) -> str:
        ":meta private:"
        return "OR"

    def visit_BoolOp(self, node: ast.BoolOp) -> str:
        ":meta private:"
        left = self.visit(node.left)
        op = self.visit(node.op)
        right = self.visit(node.right)

        # In case of a subexpression, wrap it in parentheses
        # UNLESS it has the same operator as the current BoolOp, e.g.:
        # x AND y AND z
        if isinstance(node.left, ast.BoolOp) and node.left.op != node.op:
            left = f"({left})"
        if isinstance(node.right, ast.BoolOp) and node.right.op != node.op:
            right = f"({right})"

        return f"{left} {op} {right}"

    def visit_Not(self, node: ast.Not) -> str:
        ":meta private:"
        return "NOT"

    def visit_UnaryOp(self, node: ast.UnaryOp) -> str:
        ":meta private:"
        op = self.visit(node.op)
        operand = self.visit(node.operand)

        # In case of a subexpression, wrap it in parentheses
        if isinstance(node.operand, ast.BoolOp):
            operand = f"({operand})"

        return f"{op} {operand}"

    def visit_Call(self, node: ast.Call) -> str:
        ":meta private:"
        try:
            # Grammar has already validated that the function is valid OData,
            # but that doesn't guarantee we can represent it in SQL:
            sql_gen = getattr(self, "sqlfunc_" + node.func.name.lower())
        except AttributeError:
            raise exceptions.UnsupportedFunctionException(node.func.name)

        return sql_gen(*node.args)

    def sqlfunc_concat(self, *args: ast._Node) -> str:
        ":meta private:"
        # Presto's `concat` works on both strings and lists:
        args_sql = [self.visit(arg) for arg in args]
        return f"concat({args_sql[0]}, {args_sql[1]})"

    def sqlfunc_contains(self, *args: ast._Node) -> str:
        ":meta private:"
        args_sql = [self.visit(arg) for arg in args]
        inferred_type = [typing.infer_type(arg) for arg in args]

        # If any of the inputs is a string or default, assume str-contains:
        if any(typ is ast.String for typ in inferred_type) or all(
            typ is None for typ in inferred_type
        ):
            return f"strpos({args_sql[0]}, {args_sql[1]}) > 0"

        # If any of the inputs is a list, assume list-contains:
        if any(typ is ast.List for typ in inferred_type):
            raise exceptions.UnsupportedFunctionException("contains<List>")

        raise exceptions.ArgumentTypeException("contains")

    def sqlfunc_endswith(self, *args: ast._Node) -> str:
        ":meta private:"
        args_sql = [self.visit(arg) for arg in args]
        inferred_type = [typing.infer_type(arg) for arg in args]

        # If any of the inputs is a string or default, assume str-endswith:
        if any(typ is ast.String for typ in inferred_type) or all(
            typ is None for typ in inferred_type
        ):
            return f"strpos({args_sql[0]}, {args_sql[1]}) = length({args_sql[0]}) - length({args_sql[1]}) + 1"

        # If any of the inputs is a list, assume list-endswith
        # which isn't easily doable at the moment:
        if any(typ is ast.List for typ in inferred_type):
            raise exceptions.UnsupportedFunctionException("endswith<List>")

        raise exceptions.ArgumentTypeException("endswith")

    def sqlfunc_indexof(self, *args: ast._Node) -> str:
        ":meta private:"
        args_sql = [self.visit(arg) for arg in args]
        inferred_type = [typing.infer_type(arg) for arg in args]

        # If any of the inputs is a string, assume str-indexof:
        if any(typ is ast.String for typ in inferred_type) or all(
            typ is None for typ in inferred_type
        ):
            return f"strpos({args_sql[0]}, {args_sql[1]}) - 1"

        # If any of the inputs is a list, assume list-indexof
        # which isn't easily doable at the moment:
        if any(typ is ast.List for typ in inferred_type):
            raise exceptions.UnsupportedFunctionException("indexof<List>")

        raise exceptions.ArgumentTypeException("indexof")

    def sqlfunc_length(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        inferred_type = typing.infer_type(arg)

        # If the input is a string or default, assume str-length:
        if inferred_type is ast.String or inferred_type is None:
            return f"length({arg_sql})"

        # If the input is a list, assume list-length:
        if inferred_type is ast.List:
            return f"cardinality({arg_sql})"

        raise exceptions.ArgumentTypeException("length")

    def sqlfunc_startswith(self, *args: ast._Node) -> str:
        ":meta private:"
        args_sql = [self.visit(arg) for arg in args]
        inferred_type = [typing.infer_type(arg) for arg in args]

        # If any of the inputs is a string or default, assume str-startswith:
        if any(typ is ast.String for typ in inferred_type) or all(
            typ is None for typ in inferred_type
        ):
            return f"strpos({args_sql[0]}, {args_sql[1]}) = 1"

        # If any of the inputs is a list, assume list-startswith
        # which isn't easily doable at the moment:
        if any(typ is ast.List for typ in inferred_type):
            raise exceptions.UnsupportedFunctionException("startswith<List>")

        raise exceptions.ArgumentTypeException("startswith")

    def sqlfunc_substring(self, *args: ast._Node) -> str:
        ":meta private:"
        args_sql = [self.visit(arg) for arg in args]
        inferred_type = typing.infer_type(args[0])

        # If the first input is a string or default, assume str-substr:
        if inferred_type is ast.String or inferred_type is None:
            if len(args) == 2:
                return f"substr({args_sql[0]}, {args_sql[1]} + 1)"
            if len(args) == 3:
                return f"substr({args_sql[0]}, {args_sql[1]} + 1, {args_sql[2]})"

        # If the first input is a list, assume list-substr:
        if inferred_type is ast.List:
            if len(args) == 2:
                return f"slice({args_sql[0]}, {args_sql[1]})"
            if len(args) == 3:
                return f"slice({args_sql[0]}, {args_sql[1]}, {args_sql[2]})"

        raise exceptions.ArgumentTypeException("substring")

    def sqlfunc_tolower(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"lower({arg_sql})"

    def sqlfunc_toupper(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"upper({arg_sql})"

    def sqlfunc_trim(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"trim({arg_sql})"

    def sqlfunc_year(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"year({arg_sql})"

    def sqlfunc_month(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"month({arg_sql})"

    def sqlfunc_day(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"day({arg_sql})"

    def sqlfunc_hour(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"hour({arg_sql})"

    def sqlfunc_minute(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"minute({arg_sql})"

    def sqlfunc_date(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"date_trunc(day, {arg_sql})"

    def sqlfunc_now(self) -> str:
        ":meta private:"
        return "CURRENT_TIMESTAMP"

    def sqlfunc_round(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"round({arg_sql})"

    def sqlfunc_floor(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"floor({arg_sql})"

    def sqlfunc_ceiling(self, arg: ast._Node) -> str:
        ":meta private:"
        arg_sql = self.visit(arg)
        return f"ceiling({arg_sql})"

    def sqlfunc_hassubset(self, *args: ast._Node) -> str:
        ":meta private:"
        args_sql = [self.visit(arg) for arg in args]
        return f"cardinality(array_intersect({args_sql[0]}, {args_sql[1]})) = cardinality({args_sql[1]})"



from typing import Dict

from . import ast
from .grammar import ODataLexer, ODataParser  # type: ignore
from .visitor import NodeTransformer


class AliasRewriter(NodeTransformer):
    """
    A :class:`NodeTransformer` that replaces aliases in the :term:`AST` with their
    aliased identifiers or attributes.

    Args:
        field_aliases: A mapping of aliases to their full name. These can
            be identifiers, attributes, and even function calls in odata
            syntax.
        lexer: Optional lexer instance to use. If not passed, will construct
            the default one.
        parser: Optional parser instance to use. If not passed, will construct
            the default one.
    """

    def __init__(self, field_aliases: Dict[str, str], lexer=None, parser=None):
        self.field_aliases = field_aliases

        if not lexer:
            lexer = ODataLexer()
        if not parser:
            parser = ODataParser()

        self.replacements = {
            parser.parse(lexer.tokenize(k)): parser.parse(lexer.tokenize(v))
            for k, v in self.field_aliases.items()
        }

    def visit_Identifier(self, node: ast.Identifier) -> ast._Node:
        """
        :meta private:
        """
        if node in self.replacements:
            return self.replacements[node]
        return node

    def visit_Attribute(self, node: ast.Attribute) -> ast._Node:
        """
        :meta private:
        """
        if node in self.replacements:
            return self.replacements[node]
        else:
            new_owner = self.visit(node.owner)
            return ast.Attribute(new_owner, node.attr)


class IdentifierStripper(NodeTransformer):
    """
    A :class:`NodeTransformer` that strips the given identifier off of
    attributes. E.g. ``author/name`` -> ``name``.

    Args:
        strip: The identifier to strip off of all attributes in the :term:`AST`
    """

    def __init__(self, strip: ast.Identifier):
        self.strip = strip

    def visit_Attribute(self, node: ast.Attribute) -> ast._Node:
        """
        :meta private:
        """
        if node.owner == self.strip:
            return ast.Identifier(node.attr)
        elif isinstance(node.owner, ast.Attribute):
            return ast.Attribute(self.visit(node.owner), node.attr)

        return node



import logging
import operator
from typing import Optional, Tuple, Type, Union

from . import ast, exceptions as ex

log = logging.getLogger(__name__)


def typecheck(
    node: ast._Node, expected_type: Union[Type, Tuple[Type, ...]], field_name: str
):
    """
    Checks that the inferred type of ``node`` is (one) of ``expected_type``, and
    raises :class:`ArgumentTypeException` if not.

    Args:
        node: The node to type check.
        expected_type: The allowed type(s) the node can have.
        field_name: The name of the field you're typechecking. Only used in the
            exception.
    Raises:
        ArgumentTypeException
    """
    actual_type = infer_type(node)
    compare = operator.contains if isinstance(expected_type, tuple) else operator.eq
    if actual_type and not compare(expected_type, actual_type):
        allowed = (
            [t.__name__ for t in expected_type]
            if isinstance(expected_type, tuple)
            else expected_type.__name__
        )
        raise ex.ArgumentTypeException(field_name, str(allowed), actual_type.__name__)


def infer_type(node: ast._Node) -> Optional[Type[ast._Node]]:
    """
    Tries to infer the type of ``node``.

    Args:
        node: The node to infer the type for.
    Returns:
        The inferred type or ``None`` if unable to infer.
    """
    if isinstance(node, (ast._Literal)):
        return type(node)

    if isinstance(node, (ast.Compare, ast.BoolOp)):
        return ast.Boolean

    if isinstance(node, ast.Call):
        return infer_return_type(node)

    log.warning("Failed to infer type for %s", node)
    return None


def infer_return_type(node: ast.Call) -> Optional[Type[ast._Node]]:
    """
    Tries to infer the type of a function call ``node``.

    Args:
        node: The node to infer the type for.
    Returns:
        The inferred type or ``None`` if unable to infer.
    """
    func = node.func.name

    if func in (
        "contains",
        "endswith",
        "startswith",
        "hassubset",
        "hassubsequence",
        "geo.intersects",
    ):
        return ast.Boolean

    if func in (
        "indexof",
        "length",
        "year",
        "month",
        "day",
        "hour",
        "minute",
        "second",
        "totaloffsetminutes",
    ):
        return ast.Integer

    if func in (
        "fractionalseconds",
        "totalseconds",
        "ceiling",
        "floor",
        "round",
        "geo.distance",
        "geo.length",
    ):
        return ast.Float

    if func in ("tolower", "toupper", "trim"):
        return ast.String

    if func == "date":
        return ast.Date

    if func in ("maxdatetime", "mindatetime", "now"):
        return ast.DateTime

    if func == "concat":
        return infer_type(node.args[0]) or infer_type(node.args[1])

    if func == "substring":
        return infer_type(node.args[0])

    return None



from django.db.models.query import QuerySet

from odata_query.grammar import ODataLexer, ODataParser  # type: ignore

from .django_q import AstToDjangoQVisitor


def apply_odata_query(queryset: QuerySet, odata_query: str) -> QuerySet:
    """
    Shorthand for applying an OData query to a Django QuerySet.

    Args:
        queryset: Django QuerySet to apply the OData query to.
        odata_query: OData query string.
    Returns:
        QuerySet: The modified QuerySet
    """
    lexer = ODataLexer()
    parser = ODataParser()
    model = queryset.model

    ast = parser.parse(lexer.tokenize(odata_query))
    transformer = AstToDjangoQVisitor(model)
    where_clause = transformer.visit(ast)

    if transformer.queryset_annotations:
        queryset = queryset.annotate(**transformer.queryset_annotations)

    return queryset.filter(where_clause)



from dataclasses import fields
from typing import Any, Iterator, Tuple

from . import ast


def iter_dataclass_fields(node: ast._Node) -> Iterator[Tuple[str, Any]]:
    """
    Loops over all fields of the given node, yielding the field's name and
    the current value.

    Yields:
        Tuples of ``(fieldname, value)`` for each field in ``node._fields``.
    """
    for field in fields(node):
        yield field.name, getattr(node, field.name)


class NodeVisitor:
    """
    Base class for visitors that walk the :term:`AST` and calls a visitor
    method for every node found. This method may return a value
    which is forwarded by the :func:`visit` method.

    This class is meant to be subclassed, with the subclass adding visitor
    methods.
    By default the visitor methods for the nodes are named ``'visit_'`` +
    class name of the node (e.g. ``visit_Identifier(self, identifier)``).
    If no visitor method exists for a node, the :func:`generic_visit` visitor is
    used instead.
    """

    def visit(self, node: ast._Node):
        """
        Looks for an explicit node visiting method on ``self``,
        otherwise calls :func:`generic_visit`.

        Returns:
            Whatever the called method returned. The user is free to choose what
            the :class:`NodeVisitor` should return.
        """
        method = "visit_" + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def generic_visit(self, node: ast._Node):
        """
        Visits all fields on ``node`` recursively.
        Called if no explicit visitor method exists for a node.
        """
        for field, value in iter_dataclass_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast._Node):
                        self.visit(item)
            elif isinstance(value, ast._Node):
                self.visit(value)


class NodeTransformer(NodeVisitor):
    """
    A subclass of :class:`NodeVisitor` that allows replacing of nodes in the
    :term:`AST` as it passes over it. The visitor methods should return instances
    of :class:`_Node` that replace the passed node.
    """

    def generic_visit(self, node: ast._Node) -> ast._Node:
        new_kwargs = {}

        for field, value in iter_dataclass_fields(node):
            if isinstance(value, list):
                new_val = []
                for item in value:
                    if isinstance(item, ast._Node):
                        new_val.append(self.visit(item))
                    else:
                        new_val.append(item)
            elif isinstance(value, ast._Node):
                new_val = self.visit(value)
            else:
                new_val = value

            new_kwargs[field] = new_val

        return type(node)(**new_kwargs)  # type: ignore



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['odata_query',
 'odata_query.django',
 'odata_query.sql',
 'odata_query.sqlalchemy']

package_data = \
{'': ['*']}

install_requires = \
['python-dateutil>=2.8.1,<3.0.0', 'sly>=0.4,<0.5']

extras_require = \
{'dev': ['bump2version>=1.0,<2.0'],
 'django': ['django>=2.2'],
 'docs': ['sphinx<4.0',
          'sphinx-rtd-theme>=0.5,<0.6',
          'myst-parser>=0.14,<0.15'],
 'linting': ['black==20.8b1',
             'flake8>=3.8,<4.0',
             'isort>=5.7,<6.0',
             'mypy>=0.800,<0.801',
             'vulture>=2.3,<3.0'],
 'sqlalchemy': ['sqlalchemy>=1.4,<2.0'],
 'testing': ['pytest>=6.2,<7.0', 'pytest-cov']}

setup_kwargs = {
    'name': 'odata-query',
    'version': '0.4.1',
    'description': 'An OData query parser and transpiler.',
    'long_description': '# OData-Query\n\n[![Quality Gate Status](https://sonarcloud.io/api/project_badges/measure?project=gorillaco_odata-query&metric=alert_status&token=cb35257e036d950788a0f628af7062929318482b)](https://sonarcloud.io/dashboard?id=gorillaco_odata-query)\n[![Coverage](https://sonarcloud.io/api/project_badges/measure?project=gorillaco_odata-query&metric=coverage&token=cb35257e036d950788a0f628af7062929318482b)](https://sonarcloud.io/dashboard?id=gorillaco_odata-query)\n[![Documentation Status](https://readthedocs.org/projects/odata-query/badge/?version=latest)](https://odata-query.readthedocs.io/en/latest/?badge=latest)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n`odata-query` is a library that parses [OData v4](https://www.odata.org/) filter strings, and can convert\nthem to other forms such as\n[Django Queries](https://docs.djangoproject.com/en/3.2/topics/db/queries/),\n[SQLAlchemy Queries](https://docs.sqlalchemy.org/en/14/orm/loading_objects.html),\nor just plain SQL.\n\n\n## Installation\n\n`odata-query` is available on pypi, so can be installed with the package manager\nof your choice:\n\n```bash\npip install odata-query\n# OR\npoetry add odata-query\n# OR\npipenv install odata-query\n```\n\nThe package defines the following optional `extra`s:\n\n- `django`: If you want to pin a compatible Django version.\n- `sqlalchemy`: If you want to pin a compatible SQLAlchemy version.\n\n\nThe following `extra`s relate to the development of this library:\n\n- `linting`: The linting and code style tools.\n- `testing`: Packages for running the tests.\n- `docs`: For building the project documentation.\n\n\nYou can install `extra`s by adding them between square brackets during\ninstallation:\n\n```bash\npip install odata-query[sqlalchemy]\n```\n\n## Quickstart\n\nThe most common use case is probably parsing an OData query string, and applying\nit to a query your ORM understands. For this purpose there is an all-in-one function:\n`apply_odata_query`.\n\nExample for Django:\n\n```python\nfrom odata_query.django import apply_odata_query\n\norm_query = MyModel.objects  # This can be a Manager or a QuerySet.\nodata_query = "name eq \'test\'"  # This will usually come from a query string parameter.\n\nquery = apply_odata_query(orm_query, odata_query)\nresults = query.all()\n```\n\n\nExample for SQLAlchemy:\n\n```python\nfrom odata_query.sqlalchemy import apply_odata_query\n\norm_query = select(MyModel)  # This is any form of Query or Selectable.\nodata_query = "name eq \'test\'"  # This will usually come from a query string parameter.\n\nquery = apply_odata_query(orm_query, odata_query)\nresults = session.execute(query).scalars().all()\n```\n\n<!--- splitinclude-1 -->\n\n## Advanced Usage\n\nNot all use cases are as simple as that. Luckily, `odata-query` is very modular\nand extensible. See the [Documentation](https://odata-query.readthedocs.io/en/latest)\nfor advanced usage or extending the library for other cases.\n\n<!--- splitinclude-2 -->\n\n## Contact\n\nGot any questions or ideas? We\'d love to hear from you. Check out our\n[contributing guidelines](CONTRIBUTING.md) for ways to offer feedback and\ncontribute.\n\n\n## License\n\nCopyright (c) [Gorillini NV](https://gorilla.co/).\nAll rights reserved.\n\nLicensed under the [MIT](LICENSE) License.\n',
    'author': 'Oliver Hofkens',
    'author_email': 'oliver@gorilla.co',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# type: ignore
"""
Implementation of a subset of the
`OData formal grammar <https://docs.oasis-open.org/odata/odata/v4.01/csprd06/abnf/odata-abnf-construction-rules.txt>`_ .
Implemented with `SLY <https://sly.readthedocs.io/en/latest/>`_.
"""

import re
from typing import List

from sly import Lexer, Parser

from . import ast, exceptions

_RWS = r"\s+"
_INTEGER = r"[+-]?\d+"
_DATE = r"[1-9]\d{3}-(?:0\d|1[0-2])-(?:[0-2]\d|3[01])"
_TIME = r"(?:[01]\d|2[0-3]):[0-5]\d(:?:[0-5]\d(?:\.\d{1,12})?)"

# Defines known functions and min/max nr of args:
ODATA_FUNCTIONS = {
    # String functions
    "concat": 2,
    "contains": 2,
    "endswith": 2,
    "indexof": 2,
    "length": 1,
    "startswith": 2,
    "substring": (2, 3),
    "matchesPattern": 2,
    "tolower": 1,
    "toupper": 1,
    "trim": 1,
    # Datetime functions
    "year": 1,
    "month": 1,
    "day": 1,
    "hour": 1,
    "minute": 1,
    "second": 1,
    "fractionalseconds": 1,
    "totalseconds": 1,
    "date": 1,
    "time": 1,
    "totaloffsetminutes": 1,
    "mindatetime": 0,
    "maxdatetime": 0,
    "now": 0,
    # Math functions
    "round": 1,
    "floor": 1,
    "ceiling": 1,
    # Geo functions
    "geo.distance": 1,
    "geo.length": 1,
    "geo.intersects": 2,
    # Set functions
    "hassubset": 2,
    "hassubsequence": 2,
}


class ODataLexer(Lexer):
    tokens = {
        ODATA_IDENTIFIER,
        NULL,
        STRING,
        GUID,
        DATETIME,
        DATE,
        TIME,
        DURATION,
        DECIMAL,
        INTEGER,
        BOOLEAN,
        ADD,
        SUB,
        MUL,
        DIV,
        MOD,
        UMINUS,
        AND,
        OR,
        NOT,
        EQ,
        NE,
        LT,
        LE,
        GT,
        GE,
        IN,
        ANY,
        ALL,
        WS,
    }
    literals = {"(", ")", ",", "/", ":"}
    reflags = re.I

    def error(self, token):
        """
        Error handler during tokenization

        Args:
            token: The token that failed to tokenize.
        Raises:
            TokenizingException
        """
        raise exceptions.TokenizingException(text)

    # NOTE: Ordering of tokens is important! Longer tokens preferably first

    ####################################################################################
    # Primitive literals
    ####################################################################################

    @_(r"duration'[+-]?P(?:\d+D)?(?:T(?:\d+H)?(?:\d+M)?(?:\d+(?:\.\d+)?S)?)?'")
    def DURATION(self, t):
        ":meta private:"
        val = t.value.upper()

        # Strip the prefix and single quotes:
        val = val[len("DURATION") + 1 : -1]

        t.value = ast.Duration(val)
        return t

    @_(r"'(?:[^']|'')*'")
    def STRING(self, t):
        ":meta private:"
        # Strings are single-quoted. To represent a single quote within a string,
        # double it!

        # Strip only the first and last single quote:
        val = t.value[1:-1]
        # Replace double single-quotes with a single quote:
        val = val.replace("''", "'")

        t.value = ast.String(val)
        return t

    @_(r"[\da-f]{8}-[\da-f]{4}-[\da-f]{4}-[\da-f]{4}-[\da-f]{12}")
    def GUID(self, t):
        ":meta private:"
        t.value = ast.GUID(t.value)
        return t

    @_(_DATE + r"T" + _TIME + r"?(Z|[+-](?:[01]\d|2[0-3]):[0-5]\d)?")
    def DATETIME(self, t):
        ":meta private:"
        t.value = ast.DateTime(t.value)
        return t

    @_(_DATE)
    def DATE(self, t):
        ":meta private:"
        t.value = ast.Date(t.value)
        return t

    @_(_TIME)
    def TIME(self, t):
        ":meta private:"
        t.value = ast.Time(t.value)
        return t

    @_(_INTEGER + r"((?:(?:\.\d+)(?:e[-+]?\d+))|(?:\.\d+)|(?:e[-+]?\d+))")
    def DECIMAL(self, t):
        ":meta private:"
        t.value = ast.Float(t.value)
        return t

    @_(_INTEGER)
    def INTEGER(self, t):
        ":meta private:"
        t.value = ast.Integer(t.value)
        return t

    @_(r"true|false")
    def BOOLEAN(self, t):
        ":meta private:"
        t.value = ast.Boolean(t.value)
        return t

    @_(r"null")
    def NULL(self, t):
        ":meta private:"
        t.value = ast.Null()
        return t

    ####################################################################################
    # Arithmetic
    ####################################################################################
    @_(fr"{_RWS}add{_RWS}")
    def ADD(self, t):
        ":meta private:"
        t.value = ast.Add()
        return t

    @_(fr"{_RWS}sub{_RWS}")
    def SUB(self, t):
        ":meta private:"
        t.value = ast.Sub()
        return t

    @_(fr"{_RWS}mul{_RWS}")
    def MUL(self, t):
        ":meta private:"
        t.value = ast.Mult()
        return t

    @_(fr"{_RWS}div{_RWS}")
    def DIV(self, t):
        ":meta private:"
        t.value = ast.Div()
        return t

    @_(fr"{_RWS}mod{_RWS}")
    def MOD(self, t):
        ":meta private:"
        t.value = ast.Mod()
        return t

    @_(r"-")
    def UMINUS(self, t):
        ":meta private:"
        t.value = ast.USub()
        return t

    ####################################################################################
    # Boolean logic
    ####################################################################################
    @_(fr"{_RWS}and{_RWS}")
    def AND(self, t):
        ":meta private:"
        t.value = ast.And()
        return t

    @_(fr"{_RWS}or{_RWS}")
    def OR(self, t):
        ":meta private:"
        t.value = ast.Or()
        return t

    @_(fr"not{_RWS}")
    def NOT(self, t):
        ":meta private:"
        t.value = ast.Not()
        return t

    ####################################################################################
    # Comparators
    ####################################################################################
    @_(fr"{_RWS}eq{_RWS}")
    def EQ(self, t):
        ":meta private:"
        t.value = ast.Eq()
        return t

    @_(fr"{_RWS}ne{_RWS}")
    def NE(self, t):
        ":meta private:"
        t.value = ast.NotEq()
        return t

    @_(fr"{_RWS}lt{_RWS}")
    def LT(self, t):
        ":meta private:"
        t.value = ast.Lt()
        return t

    @_(fr"{_RWS}le{_RWS}")
    def LE(self, t):
        ":meta private:"
        t.value = ast.LtE()
        return t

    @_(fr"{_RWS}gt{_RWS}")
    def GT(self, t):
        ":meta private:"
        t.value = ast.Gt()
        return t

    @_(fr"{_RWS}ge{_RWS}")
    def GE(self, t):
        ":meta private:"
        t.value = ast.GtE()
        return t

    @_(fr"{_RWS}in{_RWS}")
    def IN(self, t):
        ":meta private:"
        t.value = ast.In()
        return t

    ####################################################################################
    # Collection operators
    ####################################################################################
    @_(r"any")
    def ANY(self, t):
        ":meta private:"
        t.value = ast.Any()
        return t

    @_(r"all")
    def ALL(self, t):
        ":meta private:"
        t.value = ast.All()
        return t

    ####################################################################################
    # Misc
    ####################################################################################
    @_(r"[_a-z]\w{0,127}")
    def ODATA_IDENTIFIER(self, t):
        ":meta private:"
        t.value = ast.Identifier(t.value)
        return t

    WS = _RWS


class ODataParser(Parser):
    debugfile = None
    tokens = ODataLexer.tokens

    # Predecence from low to high.
    # See: https://docs.oasis-open.org/odata/odata/v4.0/errata03/os/complete/part2-url-conventions/odata-v4.0-errata03-os-part2-url-conventions-complete.html#_Toc453752358
    # OData 5.1.1.14 Operator Precedence
    precedence = (
        ("left", OR),
        ("left", AND),
        ("left", EQ, NE),
        ("left", GT, GE, LT, LE),
        ("left", ADD, SUB),
        ("left", MUL, DIV, MOD),
        ("right", NOT, UMINUS),
        ("left", IN),
    )

    def error(self, token):
        """
        Error handler during parsing.

        Args:
            token: The token at which point parsing failed.
        Raises:
            ParsingFailedException
        """
        eof = token is None
        raise exceptions.ParsingException(token, eof)

    @_('"(" BWS common_expr BWS ")"')
    def common_expr(self, p):
        ":meta private:"
        return p.common_expr

    @_("primitive_literal", "first_member_expr", "list_expr")
    def common_expr(self, p):
        ":meta private:"
        return p[0]

    ####################################################################################
    # Primitives
    ####################################################################################
    @_(
        "NULL",
        "INTEGER",
        "DECIMAL",
        "STRING",
        "BOOLEAN",
        "GUID",
        "DATE",
        "TIME",
        "DATETIME",
        "DURATION",
    )
    def primitive_literal(self, p):
        ":meta private:"
        return p[0]

    @_('common_expr BWS "," BWS common_expr')
    def list_items(self, p):
        ":meta private:"
        return [p[0], p[4]]

    @_('list_items BWS "," BWS common_expr')
    def list_items(self, p):
        ":meta private:"
        p.list_items.append(p.common_expr)
        return p.list_items

    @_('"(" BWS common_expr BWS "," BWS ")"')
    def list_expr(self, p):
        ":meta private:"
        # NOTE: This is NOT according to the OData standard ABNF.
        # The standard says a single item list is (x), but that collides with the parenExpr
        # rule, e.g.: concat(('a'), ('b')) could mean concat of lists or concat of str expressions.
        # Therefore we follow Python syntax: a single item list has a comma at the end.
        return ast.List([p.common_expr])

    @_('"(" BWS list_items BWS ")"')
    def list_expr(self, p):
        ":meta private:"
        return ast.List(p.list_items)

    ####################################################################################
    # Names, properties, variables
    ####################################################################################

    # NOTE: !WARNING! OData ABNF has a lot of rules with very similar names
    # in this section. Like 'propertyPath' and 'propertyPathExpr', and their
    # definitions look almost exactly the same! Stay vigilant!

    @_("member_expr")
    def first_member_expr(self, p):
        ":meta private:"
        return p.member_expr

    @_("property_path_expr")
    def member_expr(self, p):
        ":meta private:"
        return p[0]

    @_("entity_navigation_property")
    def property_path_expr(self, p):
        ":meta private:"
        return p.entity_navigation_property

    @_("entity_navigation_property single_navigation_expr")
    def property_path_expr(self, p):
        ":meta private:"
        if isinstance(p[1], ast.Attribute):
            # We want to nest attributes in the opposite direction of parsing, e.g.:
            # we prefer Attribute(Attribute(Id(A), 'created_by'), 'name')
            # over      Attribute(Id(A), Attribute(Id(created_by), 'name'))
            naive_attr = ast.Attribute(p[0], p[1])
            return self._reverse_attributes(naive_attr)
        elif isinstance(p[1], ast.CollectionLambda):
            # Very similar for CollectionLambdas:
            # We prefer the CollectionLambda to define its complete owner
            # instead of being a deeply nested attribute:
            owner: ast.Identifier = p[1].owner
            new_owner = ast.Attribute(p[0], owner.name)
            return ast.CollectionLambda(new_owner, p[1].operator, p[1].lambda_)
        else:
            return ast.Attribute(p[0], p[1].name)

    @_("ODATA_IDENTIFIER")
    def entity_navigation_property(self, p):
        ":meta private:"
        return p[0]

    @_("'/' member_expr")
    def single_navigation_expr(self, p):
        ":meta private:"
        return p.member_expr

    @_("'/' any_expr", "'/' all_expr")
    def collection_path_expr(self, p):
        ":meta private:"
        return p[1]

    @_("entity_navigation_property collection_path_expr")
    def property_path_expr(self, p):
        ":meta private:"
        return ast.CollectionLambda(p[0], *p[1])

    ####################################################################################
    # Collections
    ####################################################################################
    @_('ODATA_IDENTIFIER BWS ":" BWS common_expr')
    def lambda_(self, p):
        ":meta private:"
        return ast.Lambda(p[0], p.common_expr)

    @_('ANY "(" BWS lambda_ BWS ")"')
    def any_expr(self, p):
        ":meta private:"
        return (p[0], p.lambda_)

    @_('ANY "(" BWS ")"')
    def any_expr(self, p):
        ":meta private:"
        return (p[0], None)

    @_('ALL "(" BWS lambda_ BWS ")"')
    def all_expr(self, p):
        ":meta private:"
        return (p[0], p.lambda_)

    ####################################################################################
    # Arithmetic
    ####################################################################################
    @_("UMINUS BWS common_expr")
    def common_expr(self, p):
        ":meta private:"
        return ast.UnaryOp(p[0], p[2])

    @_(
        "common_expr ADD common_expr",
        "common_expr SUB common_expr",
        "common_expr MUL common_expr",
        "common_expr DIV common_expr",
        "common_expr MOD common_expr",
    )
    def common_expr(self, p):
        ":meta private:"
        return ast.BinOp(p[1], p[0], p[2])

    ####################################################################################
    # Comparisons
    ####################################################################################
    @_(
        "common_expr EQ common_expr",
        "common_expr NE common_expr",
        "common_expr LT common_expr",
        "common_expr LE common_expr",
        "common_expr GT common_expr",
        "common_expr GE common_expr",
        "common_expr IN list_expr",
    )
    def common_expr(self, p):
        ":meta private:"
        return ast.Compare(p[1], p[0], p[2])

    ####################################################################################
    # Boolean logic
    ####################################################################################
    @_("common_expr AND common_expr", "common_expr OR common_expr")
    def common_expr(self, p):
        ":meta private:"
        return ast.BoolOp(p[1], p[0], p[2])

    @_("NOT common_expr")
    def common_expr(self, p):
        ":meta private:"
        return ast.UnaryOp(p[0], p.common_expr)

    ####################################################################################
    # Function calls
    ####################################################################################
    def _function_call(self, func: ast.Identifier, args: List[ast._Node]):
        ":meta private:"
        func_name = func.name
        try:
            n_args_exp = ODATA_FUNCTIONS[func_name]
        except KeyError:
            raise exceptions.UnknownFunctionException(func_name)

        n_args_given = len(args)
        if isinstance(n_args_exp, int) and n_args_given != n_args_exp:
            raise exceptions.ArgumentCountException(
                func_name, n_args_exp, n_args_exp, n_args_given
            )

        if isinstance(n_args_exp, tuple) and (
            n_args_given < n_args_exp[0] or n_args_given > n_args_exp[1]
        ):
            raise exceptions.ArgumentCountException(
                func_name, n_args_exp[0], n_args_exp[1], n_args_given
            )

        return ast.Call(func, args)

    @_('ODATA_IDENTIFIER "(" ")"')
    def common_expr(self, p):
        ":meta private:"
        args = []
        return self._function_call(p[0], args)

    @_('ODATA_IDENTIFIER "(" BWS common_expr BWS ")"')
    def common_expr(self, p):
        ":meta private:"
        args = [p.common_expr]
        return self._function_call(p[0], args)

    @_("ODATA_IDENTIFIER list_expr")
    def common_expr(self, p):
        ":meta private:"
        args = p[1].val
        return self._function_call(p[0], args)

    ####################################################################################
    # Misc
    ####################################################################################
    @_("")
    def empty(self, p):
        ":meta private:"
        pass

    @_("WS")
    def BWS(self, p):
        """
        'Bad Whitespace'

        :meta private:
        """
        pass

    @_("empty")
    def BWS(self, p):
        """
        'Bad Whitespace'

        :meta private:
        """
        pass

    ####################################################################################
    # Utils
    ####################################################################################
    def _reverse_attributes(self, attr: ast.Attribute) -> ast.Attribute:
        """
        Transforms an attribute like:
        ``Attribute(Id(A), Attribute(..., 'name'))``
        into
        ``Attribute(Attribute(Id(A), ...), 'name')``

        Args:
            attr: The :class:`Attribute` to reverse.
        Returns:
            The transformed attribute.
        """
        exploded = self._explode_attr(attr)
        leaf_attr = exploded.pop()
        owner = ast.Identifier(exploded.pop(0))
        for inter in exploded:
            owner = ast.Attribute(owner, inter)

        return ast.Attribute(owner, leaf_attr)

    def _explode_attr(self, attr: ast.Attribute) -> List[str]:
        """
        Splits a (possibly nested) attribute into a list of all elements, e.g.:
        ``Attribute(Id(A), Attribute(Id(B), 'name'))``
        into
        ``[A, B, name]``

        Args:
            attr: The :class:`Attribute` to split.
        Returns:
            A list of all identifiers in the ``attr``
        """
        if isinstance(attr.owner, ast.Identifier):
            exploded = [attr.owner.name]
        elif isinstance(attr.owner, ast.Attribute):
            exploded = self._explode_attr(attr.owner)
        else:
            raise NotImplementedError()

        if isinstance(attr.attr, str):
            exploded.append(attr.attr)
        elif isinstance(attr.attr, ast.Attribute):
            exploded.extend(self._explode_attr(attr.attr))
        else:
            raise NotImplementedError

        return exploded



import datetime as dt
import operator
from typing import Any, Callable, List, Type, Union

from dateutil.parser import isoparse
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.decl_api import DeclarativeMeta
from sqlalchemy.orm.relationships import RelationshipProperty
from sqlalchemy.sql import functions
from sqlalchemy.sql.expression import (
    BinaryExpression,
    BindParameter,
    BooleanClauseList,
    ClauseElement,
    ColumnClause,
    False_,
    Null,
    True_,
    and_,
    cast,
    extract,
    false,
    literal,
    null,
    or_,
    true,
)
from sqlalchemy.types import Date, Time

from odata_query import ast, exceptions as ex, typing, utils, visitor

from . import functions_ext


class AstToSqlAlchemyClauseVisitor(visitor.NodeVisitor):
    """
    :class:`NodeVisitor` that transforms an :term:`AST` into a SQLAlchemy query
    filter clause.

    Args:
        root_model: The root model of the query.
    """

    def __init__(self, root_model: Type[DeclarativeMeta]):
        self.root_model = root_model
        self.join_relationships: List[InstrumentedAttribute] = []

    def visit_Identifier(self, node: ast.Identifier) -> ColumnClause:
        ":meta private:"
        try:
            return getattr(self.root_model, node.name)
        except AttributeError:
            raise ex.InvalidFieldException(node.name)

    def visit_Attribute(self, node: ast.Attribute) -> ColumnClause:
        ":meta private:"
        rel_attr = self.visit(node.owner)
        # Owner is an InstrumentedAttribute, hopefully of a relationship.
        # But we need the model pointed to by the relationship.
        prop_inspect = inspect(rel_attr).property
        if not isinstance(prop_inspect, RelationshipProperty):
            # TODO: new exception:
            raise ValueError(f"Not a relationship: {node.owner}")
        self.join_relationships.append(rel_attr)

        # We'd like to reference the column on the related class:
        owner_cls = prop_inspect.entity.class_
        try:
            return getattr(owner_cls, node.attr)
        except AttributeError:
            raise ex.InvalidFieldException(node.attr)

    def visit_Null(self, node: ast.Null) -> Null:
        ":meta private:"
        return null()

    def visit_Integer(self, node: ast.Integer) -> BindParameter:
        ":meta private:"
        return literal(int(node.val))

    def visit_Float(self, node: ast.Float) -> BindParameter:
        ":meta private:"
        return literal(float(node.val))

    def visit_Boolean(self, node: ast.Boolean) -> Union[True_, False_]:
        ":meta private:"
        if node.val == "true":
            return true()
        else:
            return false()

    def visit_String(self, node: ast.String) -> BindParameter:
        ":meta private:"
        return literal(node.val)

    def visit_Date(self, node: ast.Date) -> BindParameter:
        ":meta private:"
        try:
            return literal(dt.date.fromisoformat(node.val))
        except ValueError:
            raise ex.ValueException(node.val)

    def visit_DateTime(self, node: ast.DateTime) -> BindParameter:
        ":meta private:"
        try:
            return literal(isoparse(node.val))
        except ValueError:
            raise ex.ValueException(node.val)

    def visit_Time(self, node: ast.Time) -> BindParameter:
        ":meta private:"
        try:
            return literal(dt.time.fromisoformat(node.val))
        except ValueError:
            raise ex.ValueException(node.val)

    def visit_Duration(self, node: ast.Duration) -> BindParameter:
        ":meta private:"
        sign, days, hours, minutes, seconds = node.unpack()
        td = dt.timedelta(
            days=float(days or 0),
            hours=float(hours or 0),
            minutes=float(minutes or 0),
            seconds=float(seconds or 0),
        )
        if sign and sign == "-":
            td = -1 * td
        return literal(td)

    def visit_GUID(self, node: ast.GUID) -> BindParameter:
        ":meta private:"
        return literal(node.val)

    def visit_List(self, node: ast.List) -> list:
        ":meta private:"
        return [self.visit(n) for n in node.val]

    def visit_Add(self, node: ast.Add) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.add

    def visit_Sub(self, node: ast.Sub) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.sub

    def visit_Mult(self, node: ast.Mult) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.mul

    def visit_Div(self, node: ast.Div) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.truediv

    def visit_Mod(self, node: ast.Mod) -> Callable[[Any, Any], Any]:
        ":meta private:"
        return operator.mod

    def visit_BinOp(self, node: ast.BinOp) -> Any:
        ":meta private:"
        left = self.visit(node.left)
        right = self.visit(node.right)
        op = self.visit(node.op)

        return op(left, right)

    def visit_Eq(
        self, node: ast.Eq
    ) -> Callable[[ClauseElement, ClauseElement], BinaryExpression]:
        ":meta private:"
        return operator.eq

    def visit_NotEq(
        self, node: ast.NotEq
    ) -> Callable[[ClauseElement, ClauseElement], BinaryExpression]:
        ":meta private:"
        return operator.ne

    def visit_Lt(
        self, node: ast.Lt
    ) -> Callable[[ClauseElement, ClauseElement], BinaryExpression]:
        ":meta private:"
        return operator.lt

    def visit_LtE(
        self, node: ast.LtE
    ) -> Callable[[ClauseElement, ClauseElement], BinaryExpression]:
        ":meta private:"
        return operator.le

    def visit_Gt(
        self, node: ast.Gt
    ) -> Callable[[ClauseElement, ClauseElement], BinaryExpression]:
        ":meta private:"
        return operator.gt

    def visit_GtE(
        self, node: ast.GtE
    ) -> Callable[[ClauseElement, ClauseElement], BinaryExpression]:
        ":meta private:"
        return operator.ge

    def visit_In(
        self, node: ast.In
    ) -> Callable[[ClauseElement, ClauseElement], BinaryExpression]:
        ":meta private:"
        return lambda a, b: a.in_(b)

    def visit_Compare(self, node: ast.Compare) -> BinaryExpression:
        ":meta private:"
        left = self.visit(node.left)
        right = self.visit(node.right)
        op = self.visit(node.comparator)

        # If a node is a `relationship` representing a single foreign key,
        # the client meant to compare the foreign key, not the related object.
        # E.g. In "blogpost/author eq 1", left should be "blogpost/author_id"
        left = self._maybe_sub_relationship_with_foreign_key(left)
        right = self._maybe_sub_relationship_with_foreign_key(right)

        return op(left, right)

    def visit_And(
        self, node: ast.And
    ) -> Callable[[ClauseElement, ClauseElement], BooleanClauseList]:
        ":meta private:"
        return and_

    def visit_Or(
        self, node: ast.Or
    ) -> Callable[[ClauseElement, ClauseElement], BooleanClauseList]:
        ":meta private:"
        return or_

    def visit_BoolOp(self, node: ast.BoolOp) -> BooleanClauseList:
        ":meta private:"
        left = self.visit(node.left)
        right = self.visit(node.right)
        op = self.visit(node.op)
        return op(left, right)

    def visit_Not(self, node: ast.Not) -> Callable[[ClauseElement], ClauseElement]:
        ":meta private:"
        return operator.invert

    def visit_UnaryOp(self, node: ast.UnaryOp) -> ClauseElement:
        ":meta private:"
        mod = self.visit(node.op)
        val = self.visit(node.operand)

        try:
            return mod(val)
        except TypeError:
            raise ex.TypeException(node.op.__class__.__name__, val)

    def visit_Call(self, node: ast.Call) -> ClauseElement:
        ":meta private:"
        try:
            handler = getattr(self, "func_" + node.func.name.lower())
        except AttributeError:
            raise ex.UnsupportedFunctionException(node.func.name)

        return handler(*node.args)

    def visit_CollectionLambda(self, node: ast.CollectionLambda):
        ":meta private:"
        owner_prop = self.visit(node.owner)
        collection_model = inspect(owner_prop).property.entity.class_

        if node.lambda_:
            # For the lambda, we want to strip the identifier off, because
            # we will execute this as a subquery in the wanted model's context.
            subq_ast = utils.expression_relative_to_identifier(
                node.lambda_.identifier, node.lambda_.expression
            )
            subq_transformer = self.__class__(collection_model)
            subquery_filter = subq_transformer.visit(subq_ast)
        else:
            subquery_filter = None

        if isinstance(node.operator, ast.Any):
            return owner_prop.any(subquery_filter)
        else:
            # For an ALL query, invert both the filter and the EXISTS:
            if node.lambda_:
                subquery_filter = ~subquery_filter
            return ~owner_prop.any(subquery_filter)

    def func_contains(self, field: ast._Node, substr: ast._Node) -> ClauseElement:
        ":meta private:"
        return self._substr_function(field, substr, "contains")

    def func_startswith(self, field: ast._Node, substr: ast._Node) -> ClauseElement:
        ":meta private:"
        return self._substr_function(field, substr, "startswith")

    def func_endswith(self, field: ast._Node, substr: ast._Node) -> ClauseElement:
        ":meta private:"
        return self._substr_function(field, substr, "endswith")

    def func_length(self, arg: ast._Node) -> functions.Function:
        ":meta private:"
        return functions.char_length(self.visit(arg))

    def func_concat(self, *args: ast._Node) -> functions.Function:
        ":meta private:"
        return functions.concat(*[self.visit(arg) for arg in args])

    def func_indexof(self, first: ast._Node, second: ast._Node) -> functions.Function:
        ":meta private:"
        # TODO: Highly dialect dependent, might want to implement in GenericFunction:
        # Subtract 1 because OData is 0-indexed while SQL is 1-indexed
        return functions_ext.strpos(self.visit(first), self.visit(second)) - 1

    def func_substring(
        self, fullstr: ast._Node, index: ast._Node, nchars: ast._Node = None
    ) -> functions.Function:
        ":meta private:"
        # Add 1 because OData is 0-indexed while SQL is 1-indexed
        if nchars:
            return functions_ext.substr(
                self.visit(fullstr),
                self.visit(index) + 1,
                self.visit(nchars),
            )
        else:
            return functions_ext.substr(self.visit(fullstr), self.visit(index) + 1)

    def func_matchespattern(
        self, field: ast._Node, pattern: ast._Node
    ) -> functions.Function:
        ":meta private:"
        identifier = self.visit(field)
        return identifier.regexp_match(self.visit(pattern))

    def func_tolower(self, field: ast._Node) -> functions.Function:
        ":meta private:"
        return functions_ext.lower(self.visit(field))

    def func_toupper(self, field: ast._Node) -> functions.Function:
        ":meta private:"
        return functions_ext.upper(self.visit(field))

    def func_trim(self, field: ast._Node) -> functions.Function:
        ":meta private:"
        return functions_ext.ltrim(functions_ext.rtrim(self.visit(field)))

    def func_date(self, field: ast._Node) -> ClauseElement:
        ":meta private:"
        return cast(self.visit(field), Date)

    def func_day(self, field: ast._Node) -> functions.Function:
        ":meta private:"
        return extract(self.visit(field), "day")

    def func_hour(self, field: ast._Node) -> functions.Function:
        ":meta private:"
        return extract(self.visit(field), "hour")

    def func_minute(self, field: ast._Node) -> functions.Function:
        ":meta private:"
        return extract(self.visit(field), "minute")

    def func_month(self, field: ast._Node) -> functions.Function:
        ":meta private:"
        return extract(self.visit(field), "month")

    def func_now(self) -> functions.Function:
        ":meta private:"
        return functions.now()

    def func_second(self, field: ast._Node) -> functions.Function:
        ":meta private:"
        return extract(self.visit(field), "second")

    def func_time(self, field: ast._Node) -> functions.Function:
        ":meta private:"
        return cast(self.visit(field), Time)

    def func_year(self, field: ast._Node) -> functions.Function:
        ":meta private:"
        return extract(self.visit(field), "year")

    def func_ceiling(self, field: ast._Node) -> functions.Function:
        ":meta private:"
        return functions_ext.ceil(self.visit(field))

    def func_floor(self, field: ast._Node) -> functions.Function:
        ":meta private:"
        return functions_ext.floor(self.visit(field))

    def func_round(self, field: ast._Node) -> functions.Function:
        ":meta private:"
        return functions_ext.round(self.visit(field))

    def _substr_function(
        self, field: ast._Node, substr: ast._Node, func: str
    ) -> ClauseElement:
        ":meta private:"
        typing.typecheck(field, (ast.Identifier, ast.String), "field")
        typing.typecheck(substr, ast.String, "substring")

        identifier = self.visit(field)
        substring = self.visit(substr)
        op = getattr(identifier, func)

        return op(substring)

    def _maybe_sub_relationship_with_foreign_key(
        self, elem: ClauseElement
    ) -> ClauseElement:
        """
        If the given ClauseElement is a `relationship` with a single ForeignKey,
        replace it with the `ForeignKey` itself.

        :meta private:
        """
        try:
            prop_inspect = inspect(elem).property
            if isinstance(prop_inspect, RelationshipProperty):
                foreign_key = prop_inspect._calculated_foreign_keys
                if len(foreign_key) == 1:
                    return next(iter(foreign_key))
        except Exception:
            pass

        return elem
